/*-
 * Copyright (C)2023..2024 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2023..2024 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: sql.c,v 1.18 2024/04/28 23:31:22 babolo Exp $"

#define BLIN_COMPAT      4
#define MIFE_COMPAT      5
#define AAACIPA_INTERNAL 1

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#if defined(USEBDB)
# include <dbsql.h>
#elif defined(USESQLITE3)
# include <sqlite3.h>
#elif defined(USESQLITE2)
# include <sqlite.h>
#endif
#include <openssl/des.h>
#include <babolo/BLINflag.h>
#include <mife.h>
#include "aaacipa.h"

const static char qook[] =
    "PRAGMA multiversion = on;\n"
    "PRAGMA snapshot_isolation = on;\n"
    "PRAGMA auto_vacuum = INCREMENTAL;\n"
    "PRAGMA foreign_keys = 1;\n"
    "PRAGMA temp_store = 2;\n";

const static char qaaa[] =
    "SELECT \"%s\".\"%s\"\n"
    " FROM \"%s\"\n"
    " WHERE \"%s\".\"%s\" == ?\n"
    "   AND \"%s\".\"%s\" == ?\n"
    "   AND (\"%s\".\"%s\" == ? OR \"%s\".\"%s\" IS NULL)\n"
    "   AND (\"%s\".\"%s\" == ? OR \"%s\".\"%s\" IS NULL)\n"
    "ORDER BY \"%s\".\"%s\" DESC\n"
    "LIMIT 1\n"
    ";";

const static char qarams[] =
    "SELECT \"%s\".\"%s1\"\n"
    "     , \"%s\".\"%s2\"\n"
    "     , \"%s\".\"%s3\"\n"
    "     , \"%s\".\"%s4\"\n"
    "     , \"%s\".\"%s5\"\n"
    "     , \"%s\".\"%s\"\n"
    "     , \"%s\".\"%s\"\n"
    "     , \"%s\".\"%s\"\n"
    " FROM \"%s\"\n"
    "LIMIT 1\n"
    ";";

const static char qttl[] =
    "SELECT COALESCE(\"%s\".\"%s\", \"%s\".\"%s\")\n"
    "     , COALESCE(\"%s\".\"%s\", \"%s\".\"%s\")\n"
    "     , \"%s\".\"%s\"\n"
    " FROM \"%s\", \"%s\"\n"
    " WHERE \"%s\".\"%s\" == ? AND \"%s\".\"%s\" == ?\n"
    ";";

const static char qsup[] =
    "UPDATE \"%s\" SET \"%s\" = ?\n"
    "     , \"%s\" = \"%s\" + 1\n"
    " WHERE \"%s\".\"%s\" == ?\n"
    "   AND \"%s\".\"%s\" == ?\n"
    "   AND \"%s\".\"%s\" == ?\n"
    "   AND \"%s\" >= ?\n"
    "   AND (  \"%s\".\"%s\" > 0\n"
    "       OR NOT ?\n"
    "       OR (\"%s\".\"%s\" == ? AND \"%s\".\"%s\" IS ?)\n"
    "       )\n"
    ";";

const static char qins[] =
    "INSERT INTO \"%s\"(\"%s\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\")\n"
    " SELECT ?1, ?2, ?3, ?4, \"%s\".\"%s\" & 15, ?5\n"
    "  FROM \"%s\"\n"
    "  WHERE \"%s\" == ?1\n"
    ";";

const static char qhsh[] =
    "UPDATE \"%s\" SET \"%s\" = ?, \"%s\" = \"%s\"\n"
    " WHERE \"%s\" == ?\n"
    ";";

int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
aaacipa_cook(aaacipa_cf *cf) {                                                                     /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    int           exx;
    aaacipa_dbh  *db     = NULL;
    int           ex     = EX_OK;
    char         *q      ;
    int           i      ;

    ifBLIN_QX2("+");
    if  (!cf->dbhandler && !(cf->dbhandler = calloc(1, sizeof(aaacipa_dbh)))) {
        ifBLIN_QW0("calloc dbhandler");
        ex = EX_OSERR;
        goto out;
    }
    db = (aaacipa_dbh *)cf->dbhandler;
    if  (!!(ex = sqlite3_initialize())) {
        ifBLIN_QX0("Could not initialize @%d: %s(%d)", __LINE__, aaacipa_exdeco(ex), ex);
        ex = EX_SOFTWARE;
        errno = ECANCELED;
        goto out;
    }
    ex = sqlite3_open_v2( cf->dbfile
                        , &db->odf
                        , SQLITE_OPEN_READWRITE | SQLITE_OPEN_SHAREDCACHE
                        , NULL
                        );
    if  (!!ex) {
        sqlite3_shutdown();
        ifBLIN_QX0("Could not open password DB @%d: %s(%d): %s"
                  , __LINE__
                  , aaacipa_exdeco(ex)
                  , ex
                  , cf->dbfile
                  );
        ex = EX_SOFTWARE;
        errno = ECANCELED;
        goto out;
    }
    if  (!!(ex = sqlite3_extended_result_codes(db->odf, 1))) {
        ifBLIN_QX1("Could not extend result codes: %s(%d)", aaacipa_exdeco(ex), ex);
    }
    if  (!!(ex = sqlite3_exec(db->odf, qook, NULL, NULL, &q))) {
        ifBLIN_QX0("PRAGMA: %s", q);
        sqlite3_free(q);
        q = NULL;
        ex = EX_SOFTWARE;
        errno = ECANCELED;
        goto out;
    }
    ex = asprintf( &q
                 , qarams
                 , cf->cipaconf
                 , cf->fieldreo
                 , cf->cipaconf
                 , cf->fieldreo
                 , cf->cipaconf
                 , cf->fieldreo
                 , cf->cipaconf
                 , cf->fieldreo
                 , cf->cipaconf
                 , cf->fieldreo
                 , cf->cipaconf
                 , cf->fieldkey
                 , cf->cipaconf
                 , cf->fieldvinit
                 , cf->cipaconf
                 , cf->fieldttl
                 , cf->cipaconf
                 );
    if  (0 > ex) {
        ifBLIN_QW0("no mem for conf query string");
        ex = EX_OSERR;
        goto out;
    }
    if  (!!(ex = sqlite3_prepare_v2(db->odf, q, -1, &db->vmprm, NULL))) {
        ex = aaacipa_ce(db, &db->vmprm, q, __LINE__, ex, cf->dbfile, 0);
        goto out;
    }
    switch((ex = sqlite3_step(db->vmprm))) {
    case SQLITE_ROW:
        i = sqlite3_column_count(db->vmprm);
        if  (i != 8) {
            ifBLIN_QX0("Col num %d IN %s", i, cf->pop);
            ex = aaacipa_ce(db, &db->vmprm, "", __LINE__, 0, "", 0);
            goto out;
        }
        for (int k = 0; k < 5; ++k) {
            if  (SQLITE_NULL == sqlite3_column_type(db->vmprm, k)) {
                ex = aaacipa_ce(db, &db->vmprm, "reord", __LINE__, 0, "NULL", k);
                goto out;
            }
            cf->vector[k] = (u_int64_t)sqlite3_column_int64(db->vmprm, k);
        }
        if  (SQLITE_NULL == sqlite3_column_type(db->vmprm, 5)) {
            ex = aaacipa_ce(db, &db->vmprm, "key", __LINE__, 0, "", 0);
            goto out;
        }
        cf->key = (u_int64_t)sqlite3_column_int64(db->vmprm, 5);
        if  (SQLITE_NULL == sqlite3_column_type(db->vmprm, 6)) {
            ex = aaacipa_ce(db, &db->vmprm, "hash init vector", __LINE__, 0, "", 0);
            goto out;
        }
        cf->iv = (u_int64_t)sqlite3_column_int64(db->vmprm, 6);
        if  (SQLITE_NULL == sqlite3_column_type(db->vmprm, 7)) {
            ex = aaacipa_ce(db, &db->vmprm, "TTL", __LINE__, 0, "", 0);
            goto out;
        }
        cf->ttl = sqlite3_column_double(db->vmprm, 7);
        break;
    case SQLITE_DONE:
        break;
    default:
        ex = aaacipa_ce(db, &db->vmprm, "step params", __LINE__, 0, "", 0);
        goto out;
    }
    free(q);
    q = NULL;
    if  (!db->vmprm || !!(ex = sqlite3_reset(db->vmprm))) {
        ifBLIN_QW0("Could not reset @%d: %s(%d)", __LINE__, aaacipa_exdeco(ex), ex);
    }
    if  (!db->vmprm || !!(ex = sqlite3_finalize(db->vmprm))) {
        ifBLIN_QW0("Could not finalize @%d: %s(%d)", __LINE__, aaacipa_exdeco(ex), ex);
    }
    db->vmprm = NULL;
    ex = asprintf( &cf->pop
                 , qaaa
                 , cf->passwd
                 , cf->fieldcred
                 , cf->passwd
                 , cf->passwd
                 , cf->fieldcontext
                 , cf->passwd
                 , cf->fieldlogin
                 , cf->passwd
                 , cf->fieldpass
                 , cf->passwd
                 , cf->fieldpass
                 , cf->passwd
                 , cf->fieldsrcip
                 , cf->passwd
                 , cf->fieldsrcip
                 , cf->passwd
                 , cf->fieldcred
                 );
    if  (0 > ex) {
        ifBLIN_QW0("no mem for AAA query string");
        ex = EX_OSERR;
        goto out;
    }
    if  (!!(ex = sqlite3_prepare_v2(db->odf, cf->pop, -1, &db->vmaaa, NULL))) {
        ex = aaacipa_ce(db, &db->vmaaa, cf->pop, __LINE__, ex, cf->dbfile, 0);
        goto out;
    }
    ex = asprintf( &q
                 , qttl
                 , cf->sessions
                 , cf->fieldttl
                 , cf->passwd
                 , cf->fieldttl
                 , cf->sessions
                 , cf->fieldmzone
                 , cf->passwd
                 , cf->fieldmzone
                 , cf->passwd
                 , cf->fieldnick
                 , cf->sessions
                 , cf->passwd
                 , cf->sessions
                 , cf->fieldsess
                 , cf->passwd
                 , cf->fieldcred
                 );
    if  (0 > ex) {
        ifBLIN_QW0("no mem for TTL query string");
        ex = EX_OSERR;
        goto out;
    }
    if  (!!(ex = sqlite3_prepare_v2(db->odf, q, -1, &db->vmttl, NULL))) {
        ex = aaacipa_ce(db, &db->vmttl, q, __LINE__, ex, cf->dbfile, 0);
        goto out;
    }
    free(q);
    q = NULL;
    ex = asprintf( &q
                 , qsup
                 , cf->sessions
                 , cf->fieldupdt
                 , cf->fieldcnt
                 , cf->fieldcnt
                 , cf->sessions
                 , cf->fieldsess
                 , cf->sessions
                 , cf->fieldcred
                 , cf->sessions
                 , cf->fieldsessip
                 , cf->fieldupdt
                 , cf->sessions
                 , cf->fieldcred
                 , cf->sessions
                 , cf->fieldlogin
                 , cf->sessions
                 , cf->fieldpass
                 );
    if  (0 > ex) {
        ifBLIN_QW0("no mem for session update query string");
        ex = EX_OSERR;
        goto out;
    }
    if  (!!(ex = sqlite3_prepare_v2(db->odf, q, -1, &db->vmsup, NULL))) {
        ex = aaacipa_ce(db, &db->vmsup, q, __LINE__, ex, cf->dbfile, 0);
        goto out;
    }
    free(q);
    q = NULL;
    ex = asprintf( &q
                 , qins
                 , cf->sessions
                 , cf->fieldcred
                 , cf->fieldlogin
                 , cf->fieldpass
                 , cf->fieldsessip
                 , cf->fieldflags
                 , cf->fieldcreat
                 , cf->passwd
                 , cf->fieldflags
                 , cf->passwd
                 , cf->fieldcred
                 );
    if  (0 > ex) {
        ifBLIN_QW0("no mem for session install query string");
        ex = EX_OSERR;
        goto out;
    }
    if  (!!(ex = sqlite3_prepare_v2(db->odf, q, -1, &db->vmins, NULL))) {
        ex = aaacipa_ce(db, &db->vmins, q, __LINE__, ex, cf->dbfile, 0);
        goto out;
    }
    free(q);
    q = NULL;
    ex = asprintf( &q
                 , qhsh
                 , cf->sessions
                 , cf->fieldhash
                 , cf->fieldupdt
                 , cf->fieldcreat
                 , cf->fieldsess
                 );
    if  (0 > ex) {
        ifBLIN_QW0("no mem for session install query string");
        ex = EX_OSERR;
        goto out;
    }
    if  (!!(ex = sqlite3_prepare_v2(db->odf, q, -1, &db->vmhsh, NULL))) {
        ex = aaacipa_ce(db, &db->vmhsh, q, __LINE__, ex, cf->dbfile, 0);
        goto out;
    }
    free(q);
    q = NULL;
out:
    exx = ex;
    if  (!!db->vmprm || !!(ex = sqlite3_finalize(db->vmprm))) {
        ifBLIN_QW0("Could not finalize @%d: %s(%d)", __LINE__, aaacipa_exdeco(ex), ex);
    }
    db->vmprm = NULL;
    if  (!!exx) ex = exx;
    if  (!!ex) {
        aaacipa_fini(db, NULL);
        if  (!!cf->dbhandler) {
            free(cf->dbhandler);
            cf->dbhandler = NULL;
        }
        ex = exx;
    }
    ifBLIN_QX2("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
aaacipa_reset(aaacipa_cf *cf, int ln, sqlite3_stmt  *vm) {                                         /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    int ex = EX_OK;

    ifBLIN_QX2("+ %d", ln);
    if  (!!vm && !!(ex = sqlite3_clear_bindings(vm))) {
        ifBLIN_QX0("Could not clear @%d: %s(%d)", ln, aaacipa_exdeco(ex), ex);
        errno = EDOOFUS;
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (!!vm && !!(ex = sqlite3_reset(vm))) {
        ifBLIN_QX0("Could not reset @%d: %s(%d)", ln, aaacipa_exdeco(ex), ex);
        errno = EPROCUNAVAIL;
        ex = EX_SOFTWARE;
        goto out;
    }
out:
    ifBLIN_QX2("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
aaacipa_getaaa(aaacipa_cf *cf, aaacipa_pass *pass) {                                               /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    aaacipa_dbh   *db       = (aaacipa_dbh *)cf->dbhandler;
    sqlite3_stmt  *vm       = db->vmaaa;
    sqlite3_stmt **vvm      = &db->vmaaa;
    int            ex       = EX_OK;
    int            i        ;

    ifBLIN_QX2("+");
    i = 0;
    ifBLIN_QX3("bind %d", i);
    if  (!!(ex = sqlite3_bind_text(vm, ++i, pass->context, -1, SQLITE_STATIC))) {
        ex = aaacipa_ce(db, vvm, pass->context, __LINE__, ex, "Could not bind context", i);
        goto out;
    }
    ifBLIN_QX3("bind %d", i);
    if  (!!(ex = sqlite3_bind_text(vm, ++i, pass->user, -1, SQLITE_STATIC))) {
        ex = aaacipa_ce(db, vvm, pass->user, __LINE__, ex, "Could not bind user", i);
        goto out;
    }
    ifBLIN_QX3("bind %d", i);
    if  (!!(ex = sqlite3_bind_text(vm, ++i, pass->password, -1, SQLITE_STATIC))) {
        ex = aaacipa_ce(db, vvm, pass->password, __LINE__, ex, "Could not bind password", i);
        goto out;
    }
    ifBLIN_QX3("bind %d", i);
    if  (!!(ex = sqlite3_bind_text(vm, ++i, pass->srcip, -1, SQLITE_STATIC))) {
        ex = aaacipa_ce(db, vvm, pass->srcip, __LINE__, ex, "Could not bind srcip", i);
        goto out;
    }
    ifBLIN_QX3("bind %d", i);
    ex = sqlite3_step(vm);
    ifBLIN_QX3("step %d", ex);
    switch(ex) {
    case SQLITE_ROW:
        i = sqlite3_column_count(vm);
        ifBLIN_QX3("row %d", i);
        if  (i != 1) {
            ifBLIN_QX0("Col num %d IN %s", i, cf->pop);
            ex = aaacipa_ce(db, vvm, "", __LINE__, 0, "", 0);
            goto out;
        }
        if  (SQLITE_NULL != sqlite3_column_type(vm, 0)) {
            pass->cred = sqlite3_column_int64(vm, 0);
        }
        ex = EX_OK;
        break;
    case SQLITE_DONE:
        ifBLIN_QX3("done");
        ex = EX_OK;
        break;
    default:
        ex = aaacipa_ce(db, vvm, "step AAA", __LINE__, 0, "", 0);
        goto out;
    }
    if  (!!(ex = aaacipa_reset(cf, __LINE__, vm))) {
        ifBLIN_QW0("aaacipa_reset");
        goto out;
    }
out:
    ifBLIN_QX2("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
aaacipa_getttl(aaacipa_cf *cf, aaacipa_pass *pass) {                                               /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    aaacipa_dbh   *db       = (aaacipa_dbh *)cf->dbhandler;
    sqlite3_stmt  *vm       = db->vmttl;
    sqlite3_stmt **vvm      = &db->vmttl;
    int            ex       = EX_OK;
    int            i        ;

    ifBLIN_QX2("+");
    i = 0;
    if  (!!(ex = sqlite3_bind_int64(vm, ++i, (sqlite3_int64)pass->sess))) {
        ex = aaacipa_ce(db, vvm, "sess", __LINE__, ex, "Could not bind session", i);
        goto out;
    }
    if  (!!(ex = sqlite3_bind_int64(vm, ++i, pass->cred))) {
        ex = aaacipa_ce(db, vvm, "cred", __LINE__, ex, "Could not bind cred", i);
        goto out;
    }
    switch((ex = sqlite3_step(vm))) {
    case SQLITE_ROW:
        i = sqlite3_column_count(vm);
        if  (i != 3) {
            ifBLIN_QX0("Col num %d IN time query", i);
            ex = aaacipa_ce(db, vvm, "", __LINE__, 0, "", 0);
            goto out;
        }
        if  (SQLITE_NULL != sqlite3_column_type(vm, 0)) {
            pass->back = pass->now - sqlite3_column_double(vm, 0);
        } else {
            pass->back = pass->now - cf->ttl;
        }
        if  (SQLITE_NULL != sqlite3_column_type(vm, 1)) {
            pass->mzone = sqlite3_column_int(vm, 1);
        } else {
            pass->mzone = 0;
        }
        if  (SQLITE_NULL != sqlite3_column_type(vm, 2)) {
            strlcpy(pass->nick, (const char *)sqlite3_column_text(vm, 1), AAACIPA_NICKLEN + 1);
        } else {
            pass->nick[0] = '\0';
        }
        ex = EX_OK;
        break;
    case SQLITE_DONE:
        ex = EX_OK;
        break;
    default:
        ex = aaacipa_ce(db, vvm, "", __LINE__, 0, "", 0);
        goto out;
    }
    if  (!!(ex = aaacipa_reset(cf, __LINE__, vm))) {
        ifBLIN_QW0("aaacipa_reset");
        goto out;
    }
out:
    ifBLIN_QX2("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
aaacipa_getsup(aaacipa_cf *cf, aaacipa_pass *pass) {                                               /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    aaacipa_dbh   *db       = (aaacipa_dbh *)cf->dbhandler;
    sqlite3_stmt  *vm       = db->vmsup;
    sqlite3_stmt **vvm      = &db->vmsup;
    int            ex       = EX_OK;
    int            i        ;

    ifBLIN_QX2("+");
    i = 0;
    if  (!!(ex = sqlite3_bind_int64(vm, ++i, (sqlite3_int64)pass->now))) {
        ex = aaacipa_ce(db, vvm, "now", __LINE__, ex, "Could not bind now", i);
        goto out;
    }
    if  (!!(ex = sqlite3_bind_int64(vm, ++i, (sqlite3_int64)pass->sess))) {
        ex = aaacipa_ce(db, vvm, "sess", __LINE__, ex, "Could not bind session", i);
        goto out;
    }
    if  (!!(ex = sqlite3_bind_int64(vm, ++i, (sqlite3_int64)pass->cred))) {
        ex = aaacipa_ce(db, vvm, "cred", __LINE__, ex, "Could not bind cred", i);
        goto out;
    }
    if  (!!(ex = sqlite3_bind_text(vm, ++i, pass->srcip, -1, SQLITE_STATIC))) {
        ex = aaacipa_ce(db, vvm, pass->srcip, __LINE__, ex, "Could not bind srcip", i);
        goto out;
    }
    if  (!!(ex = sqlite3_bind_int64(vm, ++i, (sqlite3_int64)pass->back))) {
        ex = aaacipa_ce(db, vvm, "back", __LINE__, ex, "Could not bind back", i);
        goto out;
    }
    if  (!!(ex = sqlite3_bind_int(vm, ++i, pass->crypt))) {
        ex = aaacipa_ce(db, vvm, "crypt", __LINE__, ex, "Could not bind crypt", i);
        goto out;
    }
    if  (!!(ex = sqlite3_bind_text(vm, ++i, pass->user, -1, SQLITE_STATIC))) {
        ex = aaacipa_ce(db, vvm, pass->user, __LINE__, ex, "Could not bind user", i);
        goto out;
    }
    if  (!!(ex = sqlite3_bind_text(vm, ++i, pass->password, -1, SQLITE_STATIC))) {
        ex = aaacipa_ce(db, vvm, pass->password, __LINE__, ex, "Could not bind pass", i);
        goto out;
    }
    switch((ex = sqlite3_step(vm))) {
    case SQLITE_ROW:
        ifBLIN_QX0("Why row?");
        aaacipa_fini(db, NULL);
        ex = EX_SOFTWARE;
        errno = E2BIG;
        goto out;
    case SQLITE_DONE:
        switch(sqlite3_changes(db->odf)) {
        case 0:
            pass->sess = 0;
            break;
        case 1:
            break;
        default:
            ex = aaacipa_ce(db, vvm, "too many up", __LINE__, 0, "", 0);
            goto out;
        }
        ex = EX_OK;
        break;
    default:
        ex = aaacipa_ce(db, vvm, "step up", __LINE__, 0, "", 0);
        goto out;
    }
    if  (!!(ex = aaacipa_reset(cf, __LINE__, vm))) {
        ifBLIN_QW0("aaacipa_reset");
        goto out;
    }
out:
    ifBLIN_QX2("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
aaacipa_getins(aaacipa_cf *cf, aaacipa_pass *pass) {                                               /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    aaacipa_dbh   *db       = (aaacipa_dbh *)cf->dbhandler;
    sqlite3_stmt  *vm       = db->vmins;
    sqlite3_stmt **vvm      = &db->vmins;
    int            ex       = EX_OK;
    int            i        ;

    ifBLIN_QX2("+");
    i = 0;
    if  (!!(ex = sqlite3_bind_int64(vm, ++i, (sqlite3_int64)pass->cred))) {
        ex = aaacipa_ce(db, vvm, "cred", __LINE__, ex, "Could not bind cred", i);
        goto out;
    }
    if  (!!(ex = sqlite3_bind_text(vm, ++i, pass->user, -1, SQLITE_STATIC))) {
        ex = aaacipa_ce(db, vvm, pass->user, __LINE__, ex, "Could not bind user", i);
        goto out;
    }
    if  (!!(ex = sqlite3_bind_text(vm, ++i, pass->password, -1, SQLITE_STATIC))) {
        ex = aaacipa_ce(db, vvm, pass->password, __LINE__, ex, "Could not bind password", i);
        goto out;
    }
    if  (!!(ex = sqlite3_bind_text(vm, ++i, pass->srcip, -1, SQLITE_STATIC))) {
        ex = aaacipa_ce(db, vvm, pass->srcip, __LINE__, ex, "Could not bind srcip", i);
        goto out;
    }
    if  (!!(ex = sqlite3_bind_int64(vm, ++i, (sqlite3_int64)pass->now))) {
        ex = aaacipa_ce(db, vvm, "now", __LINE__, ex, "Could not bind now", i);
        goto out;
    }
    switch((ex = sqlite3_step(vm))) {
    case SQLITE_ROW:
        ifBLIN_QX0("Why row?");
        aaacipa_fini(db, NULL);
        ex = EX_SOFTWARE;
        errno = E2BIG;
        goto out;
    case SQLITE_DONE:
        if  (!(pass->sess = (u_int64_t)sqlite3_last_insert_rowid(db->odf))) {
            ifBLIN_QX0("NO ins");
            aaacipa_fini(db, NULL);
            ex = EX_SOFTWARE;
            errno = EDOOFUS;
            goto out;
        }
        ex = EX_OK;
        break;
    default:
        ex = aaacipa_ce(db, vvm, "No ins", __LINE__, 0, "", 0);
        goto out;
    }
out:
    ifBLIN_QX2("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
aaacipa_gethsh(aaacipa_cf *cf, aaacipa_pass *pass, const char *hash) {                             /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    aaacipa_dbh   *db       = (aaacipa_dbh *)cf->dbhandler;
    sqlite3_stmt  *vm       = db->vmhsh;
    sqlite3_stmt **vvm      = &db->vmhsh;
    int            ex       = EX_OK;
    int            i        ;

    ifBLIN_QX2("+");
    i = 0;
    if  (!!(ex = sqlite3_bind_text(vm, ++i, hash, -1, SQLITE_TRANSIENT))) {
        ex = aaacipa_ce(db, vvm, hash, __LINE__, ex, "Could not bind hash", i);
        goto out;
    }
    if  (!!(ex = sqlite3_bind_int64(vm, ++i, (sqlite3_int64)pass->sess))) {
        ex = aaacipa_ce(db, vvm, "sess", __LINE__, ex, "Could not bind session", i);
        goto out;
    }
    switch((ex = sqlite3_step(vm))) {
    case SQLITE_ROW:
        ifBLIN_QX0("Why row?");
        aaacipa_fini(db, NULL);
        ex = EX_SOFTWARE;
        errno = E2BIG;
        goto out;
    case SQLITE_DONE:
        ex = EX_OK;
        break;
    default:
        ex = aaacipa_ce(db, vvm, "No ins", __LINE__, 0, "", 0);
        goto out;
    }
out:
    ifBLIN_QX2("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
aaacipa_passq(aaacipa_cf *cf, aaacipa_pass *pass) {                                                /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    char          hash[AAACIPA_HLEN + 1];
    int           ex  = EX_OK;

    ifBLIN_QX2("+");
    if  (!!(ex = aaacipa_getaaa(cf, pass))) {
        ifBLIN_QW0("aaacipa_getaaa");
        goto out;
    }
    if  (!!(ex = aaacipa_getttl(cf, pass))) {
        ifBLIN_QW0("aaacipa_getttl");
        goto out;
    }
    if  (!!(ex = aaacipa_getsup(cf, pass))) {
        ifBLIN_QW0("aaacipa_getsup");
        goto out;
    }
    if  (!pass->sess) {
        if  (!!(ex = aaacipa_getins(cf, pass))) {
            ifBLIN_QW0("aaacipa_getins");
            goto out;
        }
        if  (!!(ex = aaacipa_endes(cf, pass->sess, hash))) {
            ifBLIN_QW0("aaacipa_endes");
            goto out;
        }
        if  (!!(ex = aaacipa_gethsh(cf, pass, hash))) {
            ifBLIN_QW0("aaacipa_gethsh");
            goto out;
    }   }
out:
    ifBLIN_QX2("- %d=%ld~", ex, pass->cred);
    return(ex);
#   undef blin_internal_flags
}
